-- table and utility functions --

local type = type
local pairs = pairs
local pcall = pcall
local xpcall = xpcall
local tostring = tostring

local str_substr = string.sub
local str_replace = string.gsub
local str_format = string.format
local str_repeat = string.rep
local str_ord = string.byte
local str_chr = string.char

local table_size = table_size
local arr_insert = table.insert
local arr_remove = table.remove
local arr_sort = table.sort
local arr_concat = table.concat
local arr_unpack = table.unpack
local arr_empty = {}

local tbl_count
do	-- prevent unneeded upvalues
	local tbl_count_size_classes = {
		["LuaCustomTable"] = true,
		["LuaTransportLine"] = true,
		["LuaFluidBox"] = true,
		["LuaInventory"] = true,
	}
	tbl_count = function(tbl)	-- use only if the type and structure of tbl is unknown at runtime
		if type(tbl) == "userdata" and tbl_count_size_classes[tbl.object_name] then
			return #tbl
		end
		return table_size(tbl)
	end
end

local function tbl_getter_func(tbl)
	return function(key)
		return tbl[key]
	end
end

local function tbl_get_raw(tbl,key)
	return tbl[key]
end

local function default_value(val,default)
	if type(val) == type(default) then return val end
	return default
end

local function tbl_set_default(tbl,key,default)
	local ret = tbl[key]
	if type(ret) == type(default) then return ret end
	tbl[key] = default
	return default
end

local function tbl_get_tbl(tbl,key)
	local ret = tbl[key]
	if type(ret) ~= "table" then
		ret = {}
		tbl[key] = ret
	end
	return ret
end

local function tbl_set_deep(tbl,...)	-- only use for arbitrary depth tables
	local t = tbl
	local keys = {...}
	local val = arr_remove(keys)
	local key = arr_remove(keys)
	for _,k in pairs(keys) do
		t = tbl_get_tbl(t,k)
	end
	t[key] = val
end

local function is_not_table(v)	-- returns false for table like objects
	local t = type(v)
	return t ~= "table" and (t ~= "userdata" or v.object_name ~= "LuaCustomTable")
end

local function tbl_copy(...)	-- table deep copy splitting references while properly mimicking recursions
	local lookup = {}
	local function tcp(tbl)
		if is_not_table(tbl) then return tbl end
		if lookup[tbl] then return lookup[tbl] end
		local ret = {}
		lookup[tbl] = ret
		for k,v in pairs(tbl) do
			ret[tcp(k)] = tcp(v)
		end
		lookup[tbl] = nil
		return ret
	end
	return tcp(...)
end

local function tbl_clone(...)	-- table deep copy properly mimicking recursions and references
	local lookup = {}
	local function tcp(tbl)
		if is_not_table(tbl) then return tbl end
		if lookup[tbl] then return lookup[tbl] end
		local ret = {}
		lookup[tbl] = ret
		for k,v in pairs(tbl) do
			ret[tcp(k)] = tcp(v)
		end
		return ret
	end
	return tcp(...)
end

local function tbl_eq(...)
	local lookup = {}
	local function teq(a,b)
		if a == b then return true end -- short circuit on the same table/value
		if is_not_table(a) or is_not_table(b) then return false end -- only if both are tables they can be equal
		if lookup[a] and lookup[a][b] or lookup[b] and lookup[b][a] then return true end -- recursive tables already under investigation are equal until proven otherwise
		if lookup[a] then -- keep lookup small
			lookup[a][b] = true
		elseif lookup[b] then -- keep lookup small
			lookup[b][a] = true
		else
			lookup[a] = {[b]=true}
		end
		for k,v in pairs(a) do
			if not teq(b[k],v) then
				return false
			end
		end
		for k,v in pairs(b) do
			if a[k] == nil and v ~= nil then
				return false
			end
		end
		return true
	end
	return teq(...)
end

local function arr_add_new(arr,val)
	for k,v in pairs(arr) do
		if tbl_eq(v,val) then return v end
	end
	arr_insert(arr,val)
	return val
end

local function arr_unique(arr)
	for a = 1,#arr-1 do
		for b = #arr,a+1,-1 do
			if tbl_eq(arr[a],arr[b]) then
				arr_remove(arr,b)
			end
		end
	end
end

local function tbl_keys(tbl,insert_func,sort_func)
	if insert_func == nil then insert_func = arr_insert end
	local arr = {}
	for k,v in pairs(tbl) do
		insert_func(arr,k)
	end
	if sort_func then
		arr_sort(arr,sort_func)
	end
	return arr
end

local function tbl_values(tbl,insert_func,sort_func)
	if insert_func == nil then insert_func = arr_insert end
	local arr = {}
	for k,v in pairs(tbl) do
		insert_func(arr,v)
	end
	if sort_func then
		arr_sort(arr,sort_func)
	end
	return arr
end

local function tbl_get(tbl,key,default)
	for k,v in pairs(tbl) do
		if tbl_eq(k,key) then
			return v
		end
	end
	return default
end

local function tbl_find_value(tbl,val)
	for k,v in pairs(tbl) do
		if tbl_eq(v,val) then
			return k,v
		end
	end
	return default
end

local function tbl_flip(tbl)
	local r = {}
	for k,v in pairs(tbl) do
		r[v] = k
	end
	return r
end

local function tbl_flip_multi(tbl)
	local r = {}
	for k,v in pairs(tbl) do
		arr_add_new(
			tbl_get_tbl(r,
				arr_add_new(tbl_keys(r,arr_add_new),v)
			)
			,k)
	end
	return r
end





-- self accelerating functions --

local get_ntt_prototype
local get_item_prototype
local get_quality_prototype
local function get_prototype(...) -- use actual prototype basetype strings
	local prototypes = prototypes
	local pt = {}
	for k,v in pairs(defines.prototypes) do
		local err,val = pcall(tbl_get_raw,prototypes,str_replace(k,"-","_"))
		if err then
			pt[k] = val
		end
	end
	get_prototype = function(typ,name) -- typ uses prototype type strings
		local tbl = pt[typ]
		return tbl and tbl[name]
	end
	-- create shorthands for frequently used types
	get_ntt_prototype = tbl_getter_func(pt["entity"])
	get_item_prototype = tbl_getter_func(pt["item"])
	get_quality_prototype = tbl_getter_func(pt["quality"])
	return get_prototype(...)
end
get_ntt_prototype = function(name)
	return get_prototype("entity",name)
end
get_item_prototype = function(name)
	return get_prototype("item",name)
end
get_quality_prototype = function(name)
	return get_prototype("quality",name)
end

local function get_prototype_stable(...) -- returned prototypes can be tested for equality
	local db = {}
	get_prototype_stable = function(typ,name)
		local r = db[typ] and db[typ][name]
		if r == nil then
			r	= get_prototype(typ,name)
			if r then
				if db[typ] then
					db[typ][name] = r
				else
					db[typ] = { [name]=r }
				end
			end
		end
		return r
	end
	return get_prototype_stable(...)
end

local function get_db(...)
	local db = storage
	if type(db) ~= "table" then
		db = {}
		storage = db
	end
	get_db = function(name)
		return tbl_get_tbl(db,name)
	end
	return get_db(...)
end

local function get_player(...)
	get_player = game.get_player
	return get_player(...)
end

local function get_inventories(ntt)
	local ret = {}
	local get = ntt.get_inventory
	for n = 1,ntt.get_max_inventory_index() do
		local i = get(n)
		if i and #i>0 then
			ret[n] = i
		end
	end
	return ret
end

local function get_first_inventory(ntt)
	local get = ntt.get_inventory
	for n = 1,ntt.get_max_inventory_index() do
		local i = get(n)
		if i and #i>0 then
			return i
		end
	end
end

local function get_inventory(ntt,num)
	if not num then return get_first_inventory(ntt) end
	local i = ntt.get_inventory(num)
	return i and #i>0 and i
end

local function get_logistic_sections(ntt)
	local logi = ntt.get_logistic_sections()
	return logi and logi.sections or arr_empty
end

local has_sortable_logistic_sections
do
	local has_sortable_logistic_sections_forbidden = defines.logistic_section_type.transitional_request_controlled
	has_sortable_logistic_sections = function(ntt)
		local logi = ntt.get_logistic_sections()
		if not logi then return false end
		logi = logi.sections
		if not logi then return false end
		for _,sect in pairs(logi) do
			if sect.type == has_sortable_logistic_sections_forbidden then return false end
		end
		return true
	end
end

local function get_player_settings_value(...)
	local get = settings.get_player_settings
	get_player_settings_value = function(plr_id,name,default)
		local var = get(plr_id)[name]
		return var and var.value or default
	end
	return get_player_settings_value(...)
end





-- debug functions --

local function dumps(...)
	local function dumpt(t)
		return tostring(t)
	end
	local mapped_id,printed_id,next_id
	local str_repl = {}
	do
		for i=0,0x1f do
			str_repl[str_chr(i)] = str_format("\\x%02x",i)
		end
		str_repl["\r"] = "\\r"
		str_repl["\n"] = "\\n"
		str_repl["\t"] = "\\t"
		str_repl["\\"] = "\\\\"
		str_repl["\""] = "\\\""
	end
	local formater = {
		["nil"] = function()
			return "nil"
		end,
		["number"] = function(v)
			return tostring(v)
		end,
		["string"] = function(v)
			return "\""..str_replace(v,"[\x00-\x1f\n\"]",str_repl).."\""
		end,
		["boolean"] = function(v)
			return v and "true" or "false"
		end,
		["function"] = dumpt,
		["thread"] = dumpt,
	}
	local function recurse(t,n,i)
		local tid = mapped_id[t] or next_id
		if tid >= next_id then
			mapped_id[t] = tid
			next_id = tid +1
		end
		if not i or printed_id[t] then return n.."#"..tid end
		printed_id[t] = true
		local ret = {n.."#"..tid}
		for k,v in pairs(t) do
			arr_insert(ret,arr_concat{
				str_repeat("\t",i),
				formater[type(k)](k),
				" => ",
				formater[type(v)](v,i+1),
			})
		end
		return arr_concat(ret,"\n")
	end
	formater["userdata"] = function(t,i)
		if t.object_name ~= "LuaCustomTable" then return dumpt(t) end
		return recurse(t,"LuaCustomTable",i)
	end
	formater["table"] = function(t,i)
		return recurse(t,"table",i)
	end
	dumps = function(...)
		mapped_id = {}
		printed_id = {}
		next_id = 0
		local ret = {}
		for _,val in pairs{...} do
			arr_insert(ret,formater[type(val)](val,1))
		end
		return arr_concat(ret,"\n")
	end
	return dumps(...)
end
local function dump(...)
	local game = game
	local print_state = {skip=defines.print_skip.never,game_state=false}
	local function print(txt)
		log(txt)
		for _,plr in pairs(game.connected_players) do
			plr.print(txt,print_state)
		end
	end
	dump = function(...)
		return print(dumps(...))
	end
	return dump(...)
end

local function safecall_err_handler(...)
-- 	return "safecall error:\n"..dumps{
-- 		msg = {...},
-- 		info = debug.getinfo(2),
-- 		trace = debug.traceback(),
-- 	}
	return arr_concat({
		"safecall error:",
		arr_concat({...},"\n"),
		debug.traceback()
	},"\n")
end

local function safecall(func,...)
	local ret = {xpcall(func,safecall_err_handler,...)}
	local suc = arr_remove(ret,1)
	if suc then return arr_unpack(ret) end
	log(arr_concat(ret,"\n"))
	for _,plr in pairs(game.connected_players) do
		plr.print{script.mod_name..".safecall-error",{"mod-name."..script.mod_name},script.mod_name}
	end
end

local function safecall_wrapper(func)
	return function(...)
		return safecall(func,...)
	end
end

-- end of utilities --









-- compare functions --

local function cmp_bp_type(...)
	local BlueprintOrder = tbl_flip{"blueprint-book","blueprint","deconstruction-planner","upgrade-planner"}
	cmp_bp_type = function(a,b)
		local x = BlueprintOrder[a] or 0
		local y = BlueprintOrder[b] or 0
		if x > y then return false end
		if x < y then return true end
		return nil
	end
	return cmp_bp_type(...)
end
local function cmp_str(a,b)
	local x = a or ""
	local y = b or ""
	if x > y then return false end
	if x < y then return true end
	return nil
end
local function cmp_comparator(...)
	local ComparatorOrder = {}
	do
		for k,v in pairs{{"<"},{"<=","≤"},{">=","≥"},{">"},{"!=","≠"},{"="}} do
			for _,c in pairs(v) do
				ComparatorOrder[c] = k
			end
		end
	end
	cmp_comparator = function(a,b)
		local x = ComparatorOrder[a] or 0
		local y = ComparatorOrder[b] or 0
		if x > y then return false end
		if x < y then return true end
		return nil
	end
	return cmp_comparator(...)
end
local function cmp_nil_check(func,a,b)
	if a then
		if b then
			return func(a,b)
		end
		return false
	end
	return b and true or nil
end
local function cmp_proto_type(x,y)
	if x.order > y.order then return false end
	if x.order < y.order then return true end
	if x.name > y.name then return false end
	if x.name < y.name then return true end
	if x.type > y.type then return false end
	if x.type < y.type then return true end
	return nil
end
local function cmp_proto(x,y)
	local c = cmp_proto_type( x.group , y.group )
	if c ~= nil then return c end
	c = cmp_proto_type( x.subgroup , y.subgroup )
	if c ~= nil then return c end
	return cmp_proto_type( x , y )
end
local function cmp_qual(x,y)
	if x.level > y.level then return false end
	if x.level < y.level then return true end
	return cmp_proto(x,y)
end
local function cmp_quality(x,y)
	return cmp_nil_check(cmp_qual,x,y)
end
local function cmp_prototype(x,y)
	return cmp_nil_check(cmp_proto,x,y)
end
local function cmp_prototype_str(...)
	local lookup = {}
	do -- nothing in here shall generate upvalues
		local protos = {}
		local order = {}
		for n,p in pairs{
			["asteroid-chunk"]="asteroid_chunk",
			["entity"]="entity",
			["equipment"]="equipment",
			["fluid"]="fluid",
			["item"]="item",
			["quality"]="quality",
			["recipe"]="recipe",
			["space-connection"]="space_connection",
			["space-location"]="space_location",
			["technology"]="technology",
			["tile"]="tile",
			["virtual"]="virtual_signal",
		} do
			lookup[n] = {}
			for k,v in pairs(prototypes[p]) do
				arr_insert(order,v)
				protos[v] = {n,k}
			end
		end
		arr_sort(order,cmp_proto)
		for k,v in pairs(order) do
			local t,n = arr_unpack(protos[v])
			lookup[t][n] = k
		end
	end -- nothing in here shall generate upvalues
	cmp_prototype_str = function(aType,aName,bType,bName)
		if not (aType and aName) then
			return bType and bName and true or nil
		end
		if not (bType and bName) then
			return false
		end
		if aType==bType and aName==bName then
			return nil
		end
		local x,y = lookup[aType][aName],lookup[bType][bName]
		if x then
			if y then
				return x < y
			end
			return false
		end
		return y and true or nil
	end
	return cmp_prototype_str(...)
end
local function cmp_quality_str(...)
	local lookup = {}
	do -- nothing in here shall generate upvalues
		local order = {}
		for _,v in pairs(prototypes["quality"]) do
			arr_insert(order,v)
		end
		arr_sort(order,cmp_qual)
		for k,v in pairs(order) do
			lookup[v.name] = k
		end
	end -- nothing in here shall generate upvalues
	cmp_quality_str = function(a,b)
		local x,y = lookup[a],lookup[b]
		if x then
			if y then
				return x < y
			end
			return false
		end
		return y and true or nil
	end
	return cmp_quality_str(...)
end
local function cmp_bp_icons(a,b)
	local x = {}
	local y = {}
	for _,v in pairs(a) do
		x[v.index] = v.signal
	end
	for _,v in pairs(b) do
		y[v.index] = v.signal
	end
	for k = 1,4 do
		local c = cmp_prototype_str(
			x[k].type or "item",x[k].name,
			y[k].type or "item",y[k].name)
		if c ~= nil then return c end
		c = cmp_quality_str(x[k].quality,y[k].quality)
		if c ~= nil then return c end
	end
end

-- sort functions --

local function sort_logistics(a,b)
	local x = a.value
	local y = b.value
	local c = cmp_prototype_str( x.type,x.name , y.type,y.name )
	if c ~= nil then return c end
	c = cmp_comparator(x.comparator,y.comparator)
	if c ~= nil then return c end
	return cmp_quality_str(x.quality,y.quality)
end
local function sort_inf_filter(a,b)
	local c = cmp_prototype_str( "item",a.name , "item",b.name )
	if c ~= nil then return c end
	c = cmp_str(a.mode,b.mode)
	if c ~= nil then return c end
	return cmp_quality_str(a.quality,b.quality)
end
local function sort_decon_filter(a,b)
	local c = cmp_prototype_str( "entity",a.name , "entity",b.name )
	if c ~= nil then return c end
	c = cmp_comparator(a.comparator,b.comparator)
	if c ~= nil then return c end
	return cmp_quality_str(a.quality,b.quality)
end
local function sort_decon_tile_filter(a,b)
	return cmp_prototype_str( "tile",a , "tile",b )
end
local function sort_upgrade_filter(a,b)
	local c = cmp_prototype_str( a.type,a.name , b.type,b.name )
	if c ~= nil then return c end
	c = cmp_comparator(a.comparator,b.comparator)
	if c ~= nil then return c end
	return cmp_quality_str(a.quality,b.quality)
end
local function sort_book_records(a,b)
	local c = cmp_bp_type(a.type,b.type)
	if c ~= nil then return c end
	return cmp_bp_icons(a.preview_icons,b.preview_icons)
end



-- configuration helper --

local cfg_has_infinity_filter = {
	["infinity-cargo-wagon"]=true,
	["infinity-container"]=true,
}
local cfg_skip_logistic_filter = {
	["rocket"]=true,
}
local cfg_inventories = {
	["asteroid-collector"]={defines.inventory.chest},
	["car"]={defines.inventory.car_trunk,defines.inventory.car_trash},
	["cargo-landing-pad"]={defines.inventory.cargo_landing_pad_main,defines.inventory.cargo_landing_pad_trash},
	["cargo-wagon"]={defines.inventory.cargo_wagon},
	["character"]={defines.inventory.character_main,defines.inventory.character_trash},
	["self"]={defines.inventory.character_main,defines.inventory.character_trash},
	["container"]={defines.inventory.chest},
	["infinity-cargo-wagon"]={defines.inventory.cargo_wagon},
	["infinity-container"]={defines.inventory.chest},
	["linked-container"]={defines.inventory.chest},
	["logistic-container"]={defines.inventory.chest,defines.inventory.logistic_container_trash},
	["proxy-container"]={defines.inventory.proxy_main},
	["roboport"]={defines.inventory.roboport_robot,defines.inventory.roboport_material},
	["rocket-silo"]={defines.inventory.rocket_silo_rocket,defines.inventory.rocket_silo_trash},
	["space-platform-hub"]={defines.inventory.hub_main,defines.inventory.hub_trash},
	["spider-vehicle"]={defines.inventory.spider_trunk,defines.inventory.spider_trash},
}
local cfg_type_map = {
	["asteroid-collector"]="chest",
	["blueprint-book"]="book",
	["car"]="car",
	["cargo-landing-pad"]="space",
	["cargo-wagon"]="train",
	["character"]="player",
	["constant-combinator"]="math",
	["container"]="chest",
	["deconstruction-planner"]="decon",
	["infinity-cargo-wagon"]="train",
	["infinity-container"]="chest",
	["linked-container"]="chest",
	["logistic-container"]="chest",
	["proxy-container"]="chest",
	["roboport"]="robo",
	["rocket-silo"]="rocket",
	["self"]="self",
	["space-platform-hub"]="space",
	["spider-vehicle"]="car",
	["upgrade-planner"]="upgrade",
}

local data_storage = {}





-- sort inventories --

local function handle_entity_inventory(ntt,num)
	local inv = cfg_inventories[ntt.type]
	if not inv then return end
	inv = inv[num]
	if not inv then return end
	inv = ntt.get_inventory(inv)
	if not inv then return end
	inv.sort_and_merge()
end

local function handle_entity_inventory_all(ntt)
	for n = 1,ntt.get_max_inventory_index() do
		local i = ntt.get_inventory(n)
		if i then
			i.sort_and_merge()
		end
	end
end

-- sort logistic section filters --

local function handle_entity_logistic_sections(ntt)
	for _,sect in pairs(get_logistic_sections(ntt)) do
		if sect.is_manual and sect.filters then
			local lst = {}
			for _,fltr in pairs(sect.filters) do
				if fltr.value then
					arr_add_new(lst,fltr)
				end
			end
			arr_sort(lst,sort_logistics)
			sect.filters = lst
		end
	end
end

-- sort infinity container filters --

local function handle_entity_infinity_filters(ntt)
	if ntt.infinity_container_filters and tbl_count(ntt.infinity_container_filters) > 1 then
		local lst = {}
		for _,fltr in pairs(ntt.infinity_container_filters) do
			if fltr.name then
				fltr.index = nil
				arr_add_new(lst,fltr)
			end
		end
		arr_sort(lst,sort_inf_filter)
		for i,fltr in pairs(lst) do
			fltr.index = i
		end
		ntt.infinity_container_filters = lst
	end
end

-- sort planner easy access --

local handle_planner_functions = {}

-- sort deconstruction planner --

handle_planner_functions["decon"] = function(_,itm)
	local lst = {}
	for n,fltr in pairs(itm.entity_filters) do
		arr_add_new(lst,fltr)
	end
	arr_sort(lst,sort_decon_filter)
 	itm.entity_filters = lst
	-- also sort tiles
	lst = {}
	for _,fltr in pairs(itm.tile_filters) do
		arr_add_new(lst,fltr)
	end
	arr_sort(lst,sort_decon_tile_filter)
	itm.tile_filters = lst
end

-- sort upgrade planner --

handle_planner_functions["upgrade"] = function(_,itm)
	local from = {}
	local to = {}
	local to_only = {}
	for i = 1,itm.mapper_count do
		local s = itm.get_mapper(i,"from")
		local d = itm.get_mapper(i,"to")
		if s and not s.name then s = nil end
		if d and not d.name then d = nil end
		if s then
			to[arr_add_new(from,s)] = d
		elseif d then
			arr_add_new(to_only,d)
		end
	end
	arr_sort(from,sort_upgrade_filter)
	arr_sort(to_only,sort_upgrade_filter)
	for i = 1,#from do
		itm.set_mapper(i,"from",from[i])
		itm.set_mapper(i,"to",to[from[i]])
	end
	for i = 1,#to_only do
		itm.set_mapper(i+#from,"from",nil)
		itm.set_mapper(i+#from,"to",to_only[i])
	end
	for i = 1+#from+#to_only,itm.mapper_count do
		itm.set_mapper(i,"from",nil)
		itm.set_mapper(i,"to",nil)
	end
end

-- sort blueprint book --

handle_planner_functions["book"] = function(plr,itm)
	if itm.object_name ~= "LuaRecord" then return end
	local cond = {}
	for _,v in pairs{"decon","upgrade","book","books"} do
		cond[v] = get_player_settings_value(plr,"sbSortedLogistics-sort-"..v,false)
	end
	local sort_fn = tbl_copy(handle_planner_functions)
	sort_fn["book"] = function(plr,book)
		if cond["books"] then
		-- recurse
			for _,rec in pairs(book.contents) do
				if rec and cond[cfg_type_map[rec.type]] then
					local f = sort_fn[cfg_type_map[rec.type]]
					if f then f(plr,rec) end
				end
			end
		end
		-- sort contents when finally available
	end
	return sort_fn["book"](plr,itm)
end









-- gui functions --

local function gui_add_setting(plr,arr,name)
	return arr_insert(arr,{
		type="checkbox",
		state=get_player_settings_value(plr,"sbSortedLogistics-sort-"..name,false),
		name="sbSortedLogistics_auto_"..name,
		caption={"mod-setting-name.sbSortedLogistics-sort-"..name},
		tooltip={"mod-setting-description.sbSortedLogistics-sort-"..name},
	})
end

local function gui_destroy(plr)
	local plr_gui = plr.gui.left["sbSortedLogistics_gui"]
	if plr_gui then
		plr_gui.destroy()
	end
end
local function gui_show(plr,elements)
	gui_destroy(plr)
	if #elements<1 then return end
	local plr_gui = plr.gui.left
	local frame = plr_gui.add{
		type="frame",
		name="sbSortedLogistics_gui",
		direction="vertical",
	}
	for _,v in pairs(elements) do
		frame.add(v)
	end
end

local function gui_show_entity(plr,ntt,typ,when)
	data_storage[plr.index] = {ntt,typ}
	local elements = {}

	if typ then
		for k,v in pairs(cfg_inventories[ntt.type] or arr_empty) do
			if get_inventory(ntt,v) and (when == "always" or not get_player_settings_value(plr,"sbSortedLogistics-sort-"..typ..k)) then
				arr_insert(elements,{
					type="button",
					name="sbSortedLogistics_sort"..k,
					caption={"sbSortedLogistics-name.btn-sort-"..typ..k},
					tooltip={"sbSortedLogistics-desc.btn-sort-"..typ..k},
				})
				if when ~= "mini" then
					gui_add_setting(plr,elements,typ..k)
				end
			end
		end
	end
	if get_first_inventory(ntt) and (when == "always" or not get_player_settings_value(plr,"sbSortedLogistics-sort-allinv")) then
		arr_insert(elements,{
			type="button",
			name="sbSortedLogistics_sorta",
			caption={"sbSortedLogistics-name.btn-sort-allinv"},
			tooltip={"sbSortedLogistics-desc.btn-sort-allinv"},
		})
		if when ~= "mini" then
			gui_add_setting(plr,elements,"allinv")
		end
	end
	if has_sortable_logistic_sections(ntt) and (when == "always" or not get_player_settings_value(plr,"sbSortedLogistics-sort-"..(typ or "alllogi"))) then
		arr_insert(elements,{
			type="button",
			name="sbSortedLogistics_sortl",
				caption={"sbSortedLogistics-name.btn-sort-"..(typ or "alllogi")},
				tooltip={"sbSortedLogistics-desc.btn-sort-"..(typ or "alllogi")},
		})
		if when ~= "mini" then
			for _,v in pairs{typ,"alllogi"} do
				gui_add_setting(plr,elements,v)
			end
		end
	end
	if cfg_has_infinity_filter[ntt.type] and (when == "always" or not get_player_settings_value(plr,"sbSortedLogistics-sort-infinity")) then
		arr_insert(elements,{
			type="button",
			name="sbSortedLogistics_sorti",
			caption={"sbSortedLogistics-name.btn-sort-infinity"},
			tooltip={"sbSortedLogistics-desc.btn-sort-infinity"},
		})
		if when ~= "mini" then
			gui_add_setting(plr,elements,"infinity")
		end
	end

	return gui_show(plr,elements)
end

local function gui_show_planner(plr,bp,typ,when)
	data_storage[plr.index] = {bp,typ}
	local elements = {}

	if when == "always" or not get_player_settings_value(plr,"sbSortedLogistics-sort-"..typ) or (typ=="book" and not get_player_settings_value(plr,"sbSortedLogistics-sort-books")) then
		if typ ~= "book" then
			arr_insert(elements,{
				type="button",
				name="sbSortedLogistics_sortp",
				caption={"sbSortedLogistics-name.btn-sort-"..typ},
				tooltip={"sbSortedLogistics-desc.btn-sort-"..typ},
			})
		end
		if when ~= "mini" then
			for _,t in pairs(typ=="book" and {"decon","upgrade"--[[,typ]],"books"} or {typ}) do
				gui_add_setting(plr,elements,t)
			end
		end
	end

	return gui_show(plr,elements)
end

local gui_click = {
	["sbSortedLogistics_sort1"] = function(_,_,obj)
		return handle_entity_inventory(obj,1)
	end,
	["sbSortedLogistics_sort2"] = function(_,_,obj)
		return handle_entity_inventory(obj,2)
	end,
	["sbSortedLogistics_sorta"] = function(_,_,obj)
		return handle_entity_inventory_all(obj)
	end,
	["sbSortedLogistics_sortl"] = function(_,_,obj)
		return handle_entity_logistic_sections(obj)
	end,
	["sbSortedLogistics_sorti"] = function(_,_,obj)
		return handle_entity_infinity_filters(obj)
	end,
	["sbSortedLogistics_sortp"] = function(_,plr,obj,typ)
		return handle_planner_functions[typ](plr,obj)
	end,
}
do
	local function gui_click_make_auto(id)
		local setting_name = "sbSortedLogistics-sort-"..id
		gui_click["sbSortedLogistics_auto_"..id] = function(el,plr)
			plr.mod_settings[setting_name] = {value=el.state}
		end
	end
	for k,_ in pairs(settings.player_default) do
		if str_substr(k,1,23) == "sbSortedLogistics-sort-" then
			gui_click_make_auto(str_substr(k,24))
		end
	end
end





-- handle gui open and auto sort

local function evt_handle_entity(plr,open,obj,typ)
	if get_player_settings_value(plr,"sbSortedLogistics-sort-allinv") then
		handle_entity_inventory_all(obj)
	elseif typ then
		for k in pairs(cfg_inventories[obj.type] or arr_empty) do
			if get_player_settings_value(plr,"sbSortedLogistics-sort-"..typ..k) then
				handle_entity_inventory(obj,k)
			end
		end
	end
	if get_player_settings_value(plr,"sbSortedLogistics-sort-alllogi") or (typ and get_player_settings_value(plr,"sbSortedLogistics-sort-"..typ)) then
		handle_entity_logistic_sections(obj)
	end
	if cfg_has_infinity_filter[obj.type] and get_player_settings_value(plr,"sbSortedLogistics-sort-infinity") then
		handle_entity_infinity_filters(obj)
	end
	local when = get_player_settings_value(plr,"sbSortedLogistics-gui-show","always")
	if open and when ~= "never" then
		return gui_show_entity(plr,obj,typ,when)
	end
end
local function evt_handle_planner(plr,open,obj,typ)
	if get_player_settings_value(plr,"sbSortedLogistics-sort-"..typ) or (typ=="book" and get_player_settings_value(plr,"sbSortedLogistics-sort-books")) then
		handle_planner_functions[typ](plr,obj)
	end
	local when = get_player_settings_value(plr,"sbSortedLogistics-gui-show","always")
	if open and when ~= "never" then
		return gui_show_planner(plr,obj,typ,when)
	end
end





-- branch out gui open close event --

local cfg_defines_controllers_character = defines.controllers.character
local evt_gui_open_close_func = {
	[defines.gui_type.controller]=function(ev,plr,open)
		local chr = plr.character
		if chr and plr.controller_type == cfg_defines_controllers_character then
			return evt_handle_entity(plr,open,chr,"self")
		end
	end,
	[defines.gui_type.entity]=function(ev,plr,open)
		local ntt = ev.entity
		return evt_handle_entity(plr,open,ntt,cfg_type_map[ntt.type])
	end,
	[defines.gui_type.item]=function(ev,plr,open)
		local bp = ev.item
		local typ = cfg_type_map[bp.name]
		if typ then
			return evt_handle_planner(plr,open,bp,typ)
		end
	end,
-- 	[defines.gui_type.blueprint_library]=function(ev,plr,open)
-- 		dump{"blueprint_library",ev}
-- 	end,
}



script.on_event(defines.events.on_gui_click, safecall_wrapper(function(pEvent)
	local fn = gui_click[pEvent.element.name]
	local dat = data_storage[pEvent.player_index]
	if fn and dat then
		return fn(pEvent.element,get_player(pEvent.player_index),arr_unpack(dat))
	end
end))
script.on_event(defines.events.on_gui_opened, safecall_wrapper(function(pEvent)
	local plr = get_player(pEvent.player_index)
	gui_destroy(plr)
	data_storage[plr.index] = nil
	local fn = evt_gui_open_close_func[pEvent.gui_type]
	if fn then return fn(pEvent,plr,true) end
end))
script.on_event(defines.events.on_gui_closed, safecall_wrapper(function(pEvent)
	local plr = get_player(pEvent.player_index)
	gui_destroy(plr)
	data_storage[plr.index] = nil
	local fn = evt_gui_open_close_func[pEvent.gui_type]
	if fn then return fn(pEvent,plr,false) end
end))
script.on_event(defines.events.on_player_cursor_stack_changed, safecall_wrapper(function(pEvent)
	local plr = get_player(pEvent.player_index)
	local fn
	local itm = plr.cursor_stack
	if itm and itm.valid_for_read then
		fn = handle_planner_functions[cfg_type_map[itm.name]]
	else
		itm = plr.cursor_record
		if itm and itm.valid_for_write then
			fn = handle_planner_functions[cfg_type_map[itm.type]]
		end
	end
	if fn then return fn(plr,itm) end
end))
